import time

import asyncio

import gmqtt
import logging


class Callbacks:

    def __init__(self):
        self.messages = []
        self.publisheds = []
        self.subscribeds = []
        self.connack = None

        self.disconnected = False
        self.connected = False

    def __str__(self):
        return str(self.messages) + str(self.messagedicts) + str(self.publisheds) + \
               str(self.subscribeds) + str(self.unsubscribeds) + str(self.disconnects)

    def clear(self):
        self.__init__()

    def on_disconnect(self, client, packet):
        logging.info('[DISCONNECTED {}]'.format(client._client_id))
        self.disconnected = True

    def on_message(self, client, topic, payload, qos, properties):
        logging.info('[RECV MSG {}] TOPIC: {} PAYLOAD: {} QOS: {} PROPERTIES: {}'
                     .format(client._client_id, topic, payload, qos, properties))
        self.messages.append((topic, payload, qos, properties))

    def on_subscribe(self, client, mid, qos, properties):
        logging.info('[SUBSCRIBED {}] QOS: {}, properties: {}'.format(client._client_id, qos, properties))
        self.subscribeds.append(mid)

    def on_connect(self, client, flags, rc, properties):
        logging.info('[CONNECTED {}]'.format(client._client_id))
        self.connected = True
        self.connack = (flags, rc, properties)

    def register_for_client(self, client):
        client.on_disconnect = self.on_disconnect
        client.on_message = self.on_message
        client.on_connect = self.on_connect
        client.on_subscribe = self.on_subscribe


async def clean_retained(host, port, username, password=None, prefix=None):
    def on_message(client, topic, payload, qos, properties):
        curclient.publish(topic, b"", qos=0, retain=True)

    curclient = gmqtt.Client(prefix + "cleanretained", clean_session=True)

    curclient.set_auth_credentials(username, password)
    curclient.on_message = on_message
    await curclient.connect(host=host, port=port)
    topic = '#' if not prefix else prefix + '#'
    curclient.subscribe(topic)
    await asyncio.sleep(10)  # wait for all retained messages to arrive
    await curclient.disconnect()
    time.sleep(.1)


async def cleanup(host, port=1883, username=None, password=None, client_ids=None, prefix=None):
    # clean all client state
    print("clean up starting")
    client_ids = client_ids or (prefix + "myclientid", prefix + "myclientid2", prefix + "myclientid3")

    for clientid in client_ids:
        curclient = gmqtt.Client(clientid.encode("utf-8"), clean_session=True)
        curclient.set_auth_credentials(username=username, password=password)
        await curclient.connect(host=host, port=port)
        time.sleep(.1)
        await curclient.disconnect()
        time.sleep(.1)

    # clean retained messages
    await clean_retained(host, port, username, password=password, prefix=prefix)
    print("clean up finished")



import struct

from .utils import unpack_variable_byte_integer, pack_variable_byte_integer, pack_utf8, unpack_utf8


class Property:
    def __init__(self, id_, bytes_struct, name, allowed_packages):
        self.id = id_
        self.bytes_struct = bytes_struct
        self.name = name
        self.allowed_packages = allowed_packages

    def loads(self, bytes_array):
        # returns dict with property name-value and remaining bytes which do not belong to this property
        if self.bytes_struct == 'u8':
            # First two bytes in UTF-8 encoded properties correspond to unicode string length
            value, left_str = unpack_utf8(bytes_array)
        elif self.bytes_struct == 'u8x2':
            value1, left_str = unpack_utf8(bytes_array)
            value2, left_str = unpack_utf8(left_str)
            value = (value1, value2)
        elif self.bytes_struct == 'b':
            str_len, = struct.unpack('!H', bytes_array[:2])
            value = bytes_array[2:2 + str_len]
            left_str = bytes_array[2 + str_len:]
        elif self.bytes_struct == 'vbi':
            value, left_str = unpack_variable_byte_integer(bytes_array)
        else:
            value, left_str = self.unpack_helper(self.bytes_struct, bytes_array)
        return {self.name: value}, left_str

    def unpack_helper(self, fmt, data):
        # unpacks property value according to format, returns value and remaining bytes
        size = struct.calcsize(fmt)
        value = struct.unpack(fmt, data[:size])
        left_str = data[size:]
        if len(value) == 1:
            value = value[0]
        return value, left_str

    def _dump_user_property(self, data, packet):
        packet.extend(struct.pack('!B', self.id))
        data1, data2 = data
        packet.extend(pack_utf8(data1))
        packet.extend(pack_utf8(data2))

    def dumps(self, data):
        # packs property value into byte array
        packet = bytearray()
        if self.bytes_struct == 'u8':
            packet.extend(struct.pack('!B', self.id))
            packet.extend(pack_utf8(data))
            return packet
        elif self.bytes_struct == 'u8x2':
            if isinstance(data[0], str):
                self._dump_user_property(data, packet)
            else:
                for kv_pair in data:
                    self._dump_user_property(kv_pair, packet)
            return packet
        elif self.bytes_struct == 'b':
            packet.extend(struct.pack('!B', self.id))
            packet.extend(struct.pack('!H', len(data)))
            packet.extend(data)
            return packet
        elif self.bytes_struct == 'vbi':
            packet.extend(struct.pack('!B', self.id))
            packet.extend(pack_variable_byte_integer(data))
            return packet
        packet.extend(struct.pack('!B', self.id))
        packet.extend(struct.pack(self.bytes_struct, data))
        return packet

    @classmethod
    def factory(cls, id_=None, name=None):
        if (name is None and id_ is None) or (name is not None and id_ is not None):
            raise ValueError('Either id or name should be not None')
        if name is not None:
            return PROPERTIES_BY_NAME.get(name)
        else:
            return PROPERTIES_BY_ID.get(id_)


PROPERTIES = [
    Property(1, '!B', 'payload_format_id', ['PUBLISH', ]),
    Property(2, '!L', 'message_expiry_interval', ['PUBLISH', ]),
    Property(3, 'u8', 'content_type', ['PUBLISH']),
    Property(8, 'u8', 'response_topic', ['PUBLISH', ]),
    Property(9, 'b', 'correlation_data', ['PUBLISH']),
    Property(11, 'vbi', 'subscription_identifier', ['PUBLISH', 'SUBSCRIBE']),
    Property(17, '!L', 'session_expiry_interval', ['CONNECT', ]),
    Property(18, 'u8', 'assigned_client_identifier', ['CONNACK', ]),
    Property(19, '!H', 'server_keep_alive', ['CONNACK']),
    Property(21, 'u8', 'auth_method', ['CONNECT', 'CONNACK', 'AUTH']),
    Property(23, '!B', 'request_problem_info', ['CONNECT']),
    Property(24, '!L', 'will_delay_interval', ['CONNECT', ]),
    Property(25, '!B', 'request_response_info', ['CONNECT']),
    Property(26, 'u8', 'response_info', ['CONNACK']),
    Property(28, 'u8', 'server_reference', ['CONNACK', 'DISCONNECT']),
    Property(31, 'u8', 'reason_string', ['CONNACK', 'PUBACK', 'PUBREC', 'PUBREL', 'PUBCOMP', 'SUBACK', 'UNSUBACK',
                                         'DISCONNECT', 'AUTH']),
    Property(33, '!H', 'receive_maximum', ['CONNECT', 'CONNACK']),
    Property(34, '!H', 'topic_alias_maximum', ['CONNECT', 'CONNACK']),
    Property(35, '!H', 'topic_alias', ['PUBLISH']),
    Property(36, '!B', 'max_qos', ['CONNACK', ]),
    Property(37, '!B', 'retain_available', ['CONNACK', ]),
    Property(38, 'u8x2', 'user_property', ['CONNECT', 'CONNACK', 'PUBLISH', 'PUBACK', 'PUBREC', 'PUBREL',
                                         'PUBCOMP', 'SUBACK', 'UNSUBACK', 'DISCONNECT', 'AUTH']),
    Property(39, '!L', 'maximum_packet_size', ['CONNECT', 'CONNACK']),
    Property(40, '!B', 'wildcard_subscription_available', ['CONNACK']),
    Property(41, '!B', 'sub_id_available', ['CONNACK', ]),
    Property(42, '!B', 'shared_subscription_available', ['CONNACK']),
]

PROPERTIES_BY_ID = {pr.id: pr for pr in PROPERTIES}
PROPERTIES_BY_NAME = {pr.name: pr for pr in PROPERTIES}





import asyncio
import os
import time
from unittest import mock

import pytest

import gmqtt
from tests.utils import Callbacks, cleanup, clean_retained

if os.getenv('TOKEN'):
    host = 'mqtt.flespi.io'
    username = os.getenv("TOKEN")
    password = None
    port = 1883
else:
    host = os.getenv("HOST", "127.0.0.1")
    username = os.getenv("USERNAME", "")
    password = os.getenv("PASSWORD", None)
    port = os.getenv("PORT", 1883)

PREFIX = 'GMQTT/' + str(time.time()) + '/'

TOPICS = (PREFIX + "TopicA",
          PREFIX + "TopicA/B",
          PREFIX + "TopicA/C",
          PREFIX + "TopicA/D",
          PREFIX + "/TopicA")
WILDTOPICS = (PREFIX + "TopicA/+",
              PREFIX + "+/C",
              PREFIX + "#",
              PREFIX + "/#",
              PREFIX + "/+",
              PREFIX + "+/+",
              PREFIX + "TopicA/#")
NOSUBSCRIBE_TOPICS = (PREFIX + "test/nosubscribe",)


@pytest.fixture()
async def init_clients():
    await cleanup(host, port, username, prefix=PREFIX)

    aclient = gmqtt.Client(PREFIX + "myclientid", clean_session=True)
    aclient.set_auth_credentials(username)
    callback = Callbacks()
    callback.register_for_client(aclient)

    bclient = gmqtt.Client(PREFIX + "myclientid2", clean_session=True)
    bclient.set_auth_credentials(username)
    callback2 = Callbacks()
    callback2.register_for_client(bclient)

    yield aclient, callback, bclient, callback2

    await aclient.disconnect()
    await bclient.disconnect()


@pytest.mark.asyncio
async def test_basic(init_clients):
    aclient, callback, bclient, callback2 = init_clients

    await aclient.connect(host=host, port=port, version=4)
    await bclient.connect(host=host, port=port, version=4)
    bclient.subscribe(TOPICS[0])
    await asyncio.sleep(1)

    aclient.publish(TOPICS[0], b"qos 0")
    aclient.publish(TOPICS[0], b"qos 1", 1)
    aclient.publish(TOPICS[0], b"qos 2", 2)
    await asyncio.sleep(1)
    assert len(callback2.messages) == 3


@pytest.mark.asyncio
async def test_basic_subscriptions(init_clients):
    aclient, callback, bclient, callback2 = init_clients

    await aclient.connect(host=host, port=port)
    await bclient.connect(host=host, port=port)

    subscriptions = [
        gmqtt.Subscription(TOPICS[1], qos=1),
        gmqtt.Subscription(TOPICS[2], qos=2),
    ]
    bclient.subscribe(subscriptions, user_property=('key', 'value'), subscription_identifier=1)

    bclient.subscribe(gmqtt.Subscription(TOPICS[3], qos=1), user_property=('key', 'value'), subscription_identifier=2)
    await asyncio.sleep(1)

    aclient.publish(TOPICS[3], b"qos 0")
    aclient.publish(TOPICS[1], b"qos 1", 1)
    aclient.publish(TOPICS[2], b"qos 2", 2)
    await asyncio.sleep(1)
    assert len(callback2.messages) == 3


@pytest.mark.asyncio
async def test_retained_message(init_clients):
    aclient, callback, bclient, callback2 = init_clients

    await aclient.connect(host=host, port=port)
    aclient.publish(TOPICS[1], b"ret qos 0", 0, retain=True, user_property=("a", "2"))
    aclient.publish(TOPICS[2], b"ret qos 1", 1, retain=True, user_property=("c", "3"))
    aclient.publish(TOPICS[3], b"ret qos 2", 2, retain=True, user_property=(("a", "2"), ("c", "3")))

    await asyncio.sleep(2)

    await bclient.connect(host=host, port=port)
    bclient.subscribe(WILDTOPICS[0], qos=2)
    await asyncio.sleep(1)

    assert len(callback2.messages) == 3
    for msg in callback2.messages:
        assert msg[3]['retain'] == True

    aclient.publish(TOPICS[2], b"ret qos 1", 1, retain=True, user_property=("c", "3"))
    await asyncio.sleep(1)

    assert len(callback2.messages) == 4
    assert callback2.messages[3][3]['retain'] == 0

    await clean_retained(host, port, username, prefix=PREFIX)


@pytest.mark.asyncio
async def test_will_message(init_clients):
    aclient, callback, bclient, callback2 = init_clients

    # re-initialize aclient with will message
    will_message = gmqtt.Message(TOPICS[2], "I'm dead finally")
    aclient = gmqtt.Client(PREFIX + "myclientid3", clean_session=True, will_message=will_message)
    aclient.set_auth_credentials(username)

    await aclient.connect(host, port=port)

    await bclient.connect(host=host, port=port)
    bclient.subscribe(TOPICS[2])

    await asyncio.sleep(1)
    await aclient.disconnect(reason_code=4)
    await asyncio.sleep(1)
    assert len(callback2.messages) == 1


@pytest.mark.asyncio
async def test_no_will_message_on_gentle_disconnect(init_clients):
    aclient, callback, bclient, callback2 = init_clients

    # re-initialize aclient with will message
    will_message = gmqtt.Message(TOPICS[2], "I'm dead finally")
    aclient = gmqtt.Client(PREFIX + "myclientid3", clean_session=True, will_message=will_message)
    aclient.set_auth_credentials(username)

    await aclient.connect(host, port=port)

    await bclient.connect(host=host, port=port)
    bclient.subscribe(TOPICS[2])

    await asyncio.sleep(1)
    await aclient.disconnect(reason_code=0)
    await asyncio.sleep(1)
    assert len(callback2.messages) == 0


@pytest.mark.asyncio
async def test_shared_subscriptions(init_clients):
    aclient, callback, bclient, callback2 = init_clients

    shared_sub_topic = '$share/sharename/{}x'.format(PREFIX)
    shared_pub_topic = PREFIX + 'x'

    await aclient.connect(host=host, port=port)
    aclient.subscribe(shared_sub_topic)
    aclient.subscribe(TOPICS[0])

    await bclient.connect(host=host, port=port)
    bclient.subscribe(shared_sub_topic)
    bclient.subscribe(TOPICS[0])

    pubclient = gmqtt.Client(PREFIX + "myclient3", clean_session=True)
    pubclient.set_auth_credentials(username)
    await pubclient.connect(host, port)

    count = 10
    for i in range(count):
        pubclient.publish(TOPICS[0], "message " + str(i), 0)
    j = 0
    while len(callback.messages) + len(callback2.messages) < 2 * count and j < 20:
        await asyncio.sleep(1)
        j += 1
    await asyncio.sleep(1)
    assert len(callback.messages) == count
    assert len(callback2.messages) == count

    callback.clear()
    callback2.clear()

    count = 10
    for i in range(count):
        pubclient.publish(shared_pub_topic, "message " + str(i), 0)
    j = 0
    while len(callback.messages) + len(callback2.messages) < count and j < 20:
        await asyncio.sleep(1)
        j += 1
    await asyncio.sleep(1)
    # Each message should only be received once
    assert len(callback.messages) + len(callback2.messages) == count
    assert len(callback.messages) > 0
    assert len(callback2.messages) > 0

    await pubclient.disconnect()


@pytest.mark.asyncio
async def test_assigned_clientid():
    noidclient = gmqtt.Client("", clean_session=True)
    noidclient.set_auth_credentials(username)
    callback = Callbacks()
    callback.register_for_client(noidclient)
    await noidclient.connect(host=host, port=port)
    await noidclient.disconnect()
    assert callback.connack[2]['assigned_client_identifier'][0] != ""


@pytest.mark.asyncio
async def test_unsubscribe(init_clients):
    aclient, callback, bclient, callback2 = init_clients
    await bclient.connect(host=host, port=port)
    await aclient.connect(host=host, port=port)

    bclient.subscribe(TOPICS[1])
    bclient.subscribe(TOPICS[2])
    bclient.subscribe(TOPICS[3])
    await asyncio.sleep(1)
    assert len(bclient.subscriptions) == 3

    aclient.publish(TOPICS[1], b"topic 0 - subscribed", 1, retain=False)
    aclient.publish(TOPICS[2], b"topic 1", 1, retain=False)
    aclient.publish(TOPICS[3], b"topic 2", 1, retain=False)
    await asyncio.sleep(1)
    assert len(callback2.messages) == 3
    callback2.clear()
    # Unsubscribe from one topic
    bclient.unsubscribe(TOPICS[1])
    assert len(bclient.subscriptions) == 2
    await asyncio.sleep(3)

    aclient.publish(TOPICS[1], b"topic 0 - unsubscribed", 1, retain=False)
    aclient.publish(TOPICS[2], b"topic 1", 1, retain=False)
    aclient.publish(TOPICS[3], b"topic 2", 1, retain=False)
    await asyncio.sleep(2)

    assert len(callback2.messages) == 2


@pytest.mark.asyncio
async def test_overlapping_subscriptions(init_clients):
    aclient, callback, bclient, callback2 = init_clients
    await bclient.connect(host=host, port=port)
    await aclient.connect(host=host, port=port)

    aclient.subscribe(TOPICS[3], qos=2, subscription_identifier=21)
    aclient.subscribe(WILDTOPICS[6], qos=1, subscription_identifier=42)
    await asyncio.sleep(1)
    bclient.publish(TOPICS[3], b"overlapping topic filters", 2)
    await asyncio.sleep(1)
    assert len(callback.messages) in [1, 2]
    if len(callback.messages) == 1:
        assert callback.messages[0][2] == 2
        assert set(callback.messages[0][3]['subscription_identifier']) == {42, 21}
    else:
        assert (callback.messages[0][2] == 2 and callback.messages[1][2] == 1) or \
               (callback.messages[0][2] == 1 and callback.messages[1][2] == 2)


@pytest.mark.asyncio
async def test_redelivery_on_reconnect(init_clients):
    # redelivery on reconnect. When a QoS 1 or 2 exchange has not been completed, the server should retry the
    # appropriate MQTT packets
    messages = []

    def on_message(client, topic, payload, qos, properties):
        print('MSG', (topic, payload, qos, properties))
        messages.append((topic, payload, qos, properties))
        return 131

    aclient, callback, bclient, callback2 = init_clients

    disconnect_client = gmqtt.Client(PREFIX + 'myclientid3', optimistic_acknowledgement=False,
                                     clean_session=False, session_expiry_interval=99999)
    disconnect_client.set_config({'reconnect_retries': 0})
    disconnect_client.on_message = on_message
    disconnect_client.set_auth_credentials(username)

    await disconnect_client.connect(host=host, port=port)
    disconnect_client.subscribe(WILDTOPICS[6], 2)

    await asyncio.sleep(1)
    await aclient.connect(host, port)
    await asyncio.sleep(1)

    aclient.publish(TOPICS[1], b"", 1, retain=False)
    aclient.publish(TOPICS[3], b"", 2, retain=False)
    await asyncio.sleep(1)
    messages = []
    await disconnect_client.reconnect()

    await asyncio.sleep(2)
    assert len(messages) == 2
    await disconnect_client.disconnect()


@pytest.mark.asyncio
async def test_async_on_message(init_clients):
    # redelivery on reconnect. When a QoS 1 or 2 exchange has not been completed, the server should retry the
    # appropriate MQTT packets
    messages = []

    async def on_message(client, topic, payload, qos, properties):
        print('MSG', (topic, payload, qos, properties))
        await asyncio.sleep(0.5)
        messages.append((topic, payload, qos, properties))
        return 131

    aclient, callback, bclient, callback2 = init_clients

    disconnect_client = gmqtt.Client(PREFIX + 'myclientid3', optimistic_acknowledgement=False,
                                     clean_session=False, session_expiry_interval=99999)
    disconnect_client.set_config({'reconnect_retries': 0})
    disconnect_client.on_message = on_message
    disconnect_client.set_auth_credentials(username)

    await disconnect_client.connect(host=host, port=port)
    disconnect_client.subscribe(WILDTOPICS[6], 2)

    await asyncio.sleep(1)
    await aclient.connect(host, port)
    await asyncio.sleep(1)

    aclient.publish(TOPICS[1], b"", 1, retain=False)
    aclient.publish(TOPICS[3], b"", 2, retain=False)
    await asyncio.sleep(2)
    messages = []
    await disconnect_client.reconnect()

    await asyncio.sleep(2)
    assert len(messages) == 2
    await disconnect_client.disconnect()


@pytest.mark.asyncio
async def test_request_response(init_clients):
    aclient, callback, bclient, callback2 = init_clients

    await aclient.connect(host=host, port=port)
    await bclient.connect(host=host, port=port)

    aclient.subscribe(WILDTOPICS[0], 2)

    bclient.subscribe(WILDTOPICS[0], 2)

    await asyncio.sleep(1)
    # client a is the requester
    aclient.publish(TOPICS[1], b"request", 1, response_topic=TOPICS[2], correlation_data=b'334')

    await asyncio.sleep(1)

    # client b is the responder
    assert len(callback2.messages) == 1

    assert callback2.messages[0][3]['response_topic'] == [TOPICS[2], ]
    assert callback2.messages[0][3]['correlation_data'] == [b"334", ]

    bclient.publish(callback2.messages[0][3]['response_topic'][0], b"response", 1,
                    correlation_data=callback2.messages[0][3]['correlation_data'][0])

    await asyncio.sleep(1)
    assert len(callback.messages) == 2


@pytest.mark.asyncio
async def test_subscribe_no_local(init_clients):
    aclient, callback, bclient, callback2 = init_clients

    await aclient.connect(host=host, port=port)
    await bclient.connect(host=host, port=port)

    aclient.subscribe(WILDTOPICS[0], 2, no_local=True)

    bclient.subscribe(WILDTOPICS[0], 2)

    await asyncio.sleep(1)

    aclient.publish(TOPICS[1], b"aclient msg", 1)

    bclient.publish(TOPICS[1], b"bclient msg", 1)

    await asyncio.sleep(1)

    assert len(callback.messages) == 1
    assert len(callback2.messages) == 2


@pytest.mark.asyncio
async def test_subscribe_retain_01_handling_flag(init_clients):
    aclient, callback, bclient, callback2 = init_clients

    await aclient.connect(host=host, port=port)
    await bclient.connect(host=host, port=port)

    aclient.publish(TOPICS[1], b"ret qos 1", 1, retain=True)

    await asyncio.sleep(1)

    bclient.subscribe(WILDTOPICS[0], qos=2, retain_handling_options=0)

    await asyncio.sleep(1)

    assert len(callback2.messages) == 1

    bclient.subscribe(WILDTOPICS[0], qos=2, retain_handling_options=0)

    await asyncio.sleep(1)

    assert len(callback2.messages) == 2

    bclient.subscribe(WILDTOPICS[0], qos=2, retain_handling_options=1)

    await asyncio.sleep(1)

    assert len(callback2.messages) == 2


@pytest.mark.asyncio
async def test_subscribe_retain_2_handling_flag(init_clients):
    aclient, callback, bclient, callback2 = init_clients

    await aclient.connect(host=host, port=port)
    await bclient.connect(host=host, port=port)

    aclient.publish(TOPICS[1], b"ret qos 1", 1, retain=True)

    await asyncio.sleep(1)

    bclient.subscribe(WILDTOPICS[0], qos=2, retain_handling_options=2)

    await asyncio.sleep(1)

    assert len(callback2.messages) == 0


@pytest.mark.asyncio
async def test_basic_ssl(init_clients):
    aclient, callback, bclient, callback2 = init_clients
    ssl_port = 8883

    await aclient.connect(host=host, port=ssl_port, ssl=True, version=4)
    await bclient.connect(host=host, port=ssl_port, ssl=True, version=5)
    bclient.subscribe(TOPICS[0])
    await asyncio.sleep(1)

    aclient.publish(TOPICS[0], b"qos 0")
    aclient.publish(TOPICS[0], b"qos 1", 1)
    aclient.publish(TOPICS[0], b"qos 2", 2)
    await asyncio.sleep(1)
    assert len(callback2.messages) == 3


@pytest.mark.asyncio
async def test_reconnection_with_failure(init_clients):
    aclient, callback, bclient, callback2 = init_clients
    aclient.set_config({'reconnect_retries': -1, 'reconnect_delay': 0})
    await aclient.connect(host=host, port=port)
    await bclient.connect(host=host, port=port)

    bclient.subscribe(TOPICS[0])

    with mock.patch.object(aclient, '_disconnect') as disconnect_mock:
        disconnect_mock.side_effect = ConnectionAbortedError("error")
        await aclient.reconnect()

    await asyncio.sleep(3)

    # Check aclient is still working after reconnection
    aclient.publish(TOPICS[0], b"test")
    await asyncio.sleep(3)
    assert len(callback2.messages) == 1



#!/usr/bin/python3
import sys
from os import path

from setuptools import find_packages, setup

import gmqtt

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as readme:
    long_description = readme.read()

extra = {}
if sys.version_info >= (3, 4):
    extra["use_2to3"] = False
    extra["convert_2to3_doctests"] = ["README.md"]

CLASSIFIERS = [
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: MIT License",
    "Natural Language :: English",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Topic :: Software Development :: Libraries :: Python Modules",
]

KEYWORDS = "Gurtam MQTT client."

TESTS_REQUIRE = [
    "atomicwrites>=1.3.0",
    "attrs>=19.1.0",
    "codecov>=2.0.15",
    "coverage>=4.5.3",
    "more-itertools>=7.0.0",
    "pluggy>=0.11.0",
    "py>=1.8.0",
    "pytest-asyncio>=0.12.0",
    "pytest-cov>=2.7.1",
    "pytest>=5.4.0",
    "six>=1.12.0",
    "uvloop>=0.14.0",
]

# Allow you to run pip install .[test] to get test dependencies included
EXTRAS_REQUIRE = {"test": TESTS_REQUIRE}

setup(
    name="gmqtt",
    version=gmqtt.__version__,
    description="Client for MQTT protocol",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=gmqtt.__author__,
    author_email=gmqtt.__email__,
    license='MIT',
    url="https://github.com/wialon/gmqtt",
    packages=find_packages(exclude=['examples', 'tests']),
    download_url="https://github.com/wialon/gmqtt",
    classifiers=CLASSIFIERS,
    keywords=KEYWORDS,
    zip_safe=True,
    test_suite="tests",
    install_requires=[],
    tests_require=TESTS_REQUIRE,
    extras_require=EXTRAS_REQUIRE,
    python_requires='>=3.5',
)



import logging
import os
import signal
import time

import uvloop
import asyncio
import gmqtt

asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

STOP = asyncio.Event()


def on_connect(client, flags, rc, properties):
    logging.info('[CONNECTED {}]'.format(client._client_id))


def on_message(client, topic, payload, qos, properties):
    logging.info('[RECV MSG {}] TOPIC: {} PAYLOAD: {} QOS: {} PROPERTIES: {}'
                 .format(client._client_id, topic, payload, qos, properties))


def on_disconnect(client, packet, exc=None):
    logging.info('[DISCONNECTED {}]'.format(client._client_id))


def on_subscribe(client, mid, qos, properties):
    logging.info('[SUBSCRIBED {}] QOS: {}'.format(client._client_id, qos))


def assign_callbacks_to_client(client):
    client.on_connect = on_connect
    client.on_message = on_message
    client.on_disconnect = on_disconnect
    client.on_subscribe = on_subscribe


def ask_exit(*args):
    STOP.set()


async def main(broker_host, broker_port, token):
    # initiate first client subscribed to TEST/SHARED/# in group mytestgroup
    sub_clientA = gmqtt.Client("clientgonnasubA")

    assign_callbacks_to_client(sub_clientA)
    sub_clientA.set_auth_credentials(token, None)
    await sub_clientA.connect(broker_host, broker_port)

    sub_clientA.subscribe('$share/mytestgroup/TEST/SHARED/#')

    # initiate second client subscribed to TEST/SHARED/# in group mytestgroup
    sub_clientB = gmqtt.Client("clientgonnasubB")

    assign_callbacks_to_client(sub_clientB)
    sub_clientB.set_auth_credentials(token, None)
    await sub_clientB.connect(broker_host, broker_port)

    sub_clientB.subscribe('$share/mytestgroup/TEST/SHARED/#')

    # this client will publish messages to TEST/SHARED/... topics
    pub_client = gmqtt.Client("clientgonnapub")

    assign_callbacks_to_client(pub_client)
    pub_client.set_auth_credentials(token, None)
    await pub_client.connect(broker_host, broker_port)

    # some of this messages will be received by client sub_clientA,
    # and another part by client sub_clientB, approximately 50/50
    for i in range(100):
        pub_client.publish('TEST/SHARED/{}'.format(i), i, user_property=('time', str(time.time())))

    await STOP.wait()
    await pub_client.disconnect()
    await sub_clientA.disconnect(session_expiry_interval=0)
    await sub_clientB.disconnect(session_expiry_interval=0)


if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    logging.basicConfig(level=logging.INFO)

    host = os.environ.get('HOST', 'mqtt.flespi.io')
    port = 1883
    token = os.environ.get('TOKEN', 'fake token')

    loop.add_signal_handler(signal.SIGINT, ask_exit)
    loop.add_signal_handler(signal.SIGTERM, ask_exit)

    loop.run_until_complete(main(host, port, token))
